#ifndef PREDICATEDEFAULTSPEC_H
#define PREDICATEDEFAULTSPEC_H

namespace PsimagLite
{

class PredicateDefaultSpec
{

public:

	void operator()(PsimagLite::String) const { }
};
} // namespace PsimagLite
#endif // PREDICATEDEFAULTSPEC_H
