#ifndef ANSICOLORS_H
#define ANSICOLORS_H
#include "AllocatorCpu.h"

namespace PsimagLite
{

namespace AnsiColor
{

	static const String red = "\x1b[31m";

	static const String green = "\x1b[32m";

	static const String yellow = "\x1b[33m";

	static const String blue = "\x1b[34m";

	static const String magenta = "\x1b[35m";

	static const String cyan = "\x1b[36m";

	static const String reset = "\x1b[0m";
} // namespace AnsiColor

} // namespace PsimagLite
#endif // ANSICOLORS_H
